from blends.ctx import ctx
from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
    match_ast_group_d,
)
from blends.syntax.builders.class_decl import (
    build_class_node,
)
from blends.syntax.builders.method_declaration import (
    DirectChildren,
    ListChildren,
    build_method_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def _is_classmethod_decorator(args: SyntaxGraphArgs, dec_id: NId) -> bool:
    graph = args.ast_graph
    childs = adj_ast(graph, dec_id)
    if not childs:
        return False
    return node_to_str(graph, childs[-1]) == "classmethod"


def _build_with_classmethod_flag(
    args: SyntaxGraphArgs,
    modifiers_list: list[NId],
    name: str,
    direct: DirectChildren,
    list_children: ListChildren,
) -> NId:
    use_flag = any(
        _is_classmethod_decorator(args, dec_id) for dec_id in modifiers_list
    ) and ctx.has_feature_flag("StackGraph")
    prior = args.metadata.get("is_classmethod")
    if use_flag:
        args.metadata["is_classmethod"] = True
    try:
        return build_method_declaration_node(args, name, direct, list_children)
    finally:
        if use_flag:
            if prior is None:
                args.metadata.pop("is_classmethod", None)
            else:
                args.metadata["is_classmethod"] = prior


def _build_function_node(args: SyntaxGraphArgs, def_id: NId, modifiers_list: list[NId]) -> NId:
    graph = args.ast_graph
    method_node = graph.nodes[def_id]
    name = node_to_str(graph, method_node["label_field_name"])
    block_id = method_node["label_field_body"]
    parameters_id = method_node["label_field_parameters"]
    if "__0__" not in match_ast(graph, parameters_id, "(", ")"):
        parameters_id = None

    direct: DirectChildren = {"block_id": block_id}
    if parameters_id:
        direct["parameters_id"] = parameters_id

    list_children: ListChildren = {}
    if modifiers_list:
        list_children["modifiers_ids"] = modifiers_list

    return _build_with_classmethod_flag(args, modifiers_list, name, direct, list_children)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    def_id = n_attrs["label_field_definition"]
    modifiers_list = match_ast_group_d(graph, args.n_id, "decorator")

    if graph.nodes[def_id]["label_type"] == "function_definition":
        return _build_function_node(args, def_id, modifiers_list)

    class_node = graph.nodes[def_id]
    name = node_to_str(graph, class_node["label_field_name"])
    block_id = class_node["label_field_body"]
    return build_class_node(args, name, block_id, modifiers_list)



from blends.syntax.models import (
    Dispatcher,
)
from blends.syntax.readers.python import (
    argument as python_argument,
)
from blends.syntax.readers.python import (
    argument_list as python_argument_list,
)
from blends.syntax.readers.python import (
    array_node as python_array,
)
from blends.syntax.readers.python import (
    as_pattern as python_as_pattern,
)
from blends.syntax.readers.python import (
    assert_statement as python_assert_statement,
)
from blends.syntax.readers.python import (
    assignment as python_assignment,
)
from blends.syntax.readers.python import (
    attribute as python_attribute,
)
from blends.syntax.readers.python import (
    await_expression as python_await_expression,
)
from blends.syntax.readers.python import (
    binary_expression as python_binary_expression,
)
from blends.syntax.readers.python import (
    boolean_literal as python_boolean_literal,
)
from blends.syntax.readers.python import (
    break_statement as python_break_statement,
)
from blends.syntax.readers.python import (
    call as python_call,
)
from blends.syntax.readers.python import (
    class_definition as python_class_definition,
)
from blends.syntax.readers.python import (
    comment as python_comment,
)
from blends.syntax.readers.python import (
    comprehension as python_comprehension,
)
from blends.syntax.readers.python import (
    conditional_expression as python_conditional_expression,
)
from blends.syntax.readers.python import (
    continue_statement as python_continue_statement,
)
from blends.syntax.readers.python import (
    decorated_definition as python_decorated_definition,
)
from blends.syntax.readers.python import (
    decorator as python_decorator,
)
from blends.syntax.readers.python import (
    dictionary as python_dictionary,
)
from blends.syntax.readers.python import (
    dictionary_splat as python_dictionary_splat,
)
from blends.syntax.readers.python import (
    else_clause as python_else_clause,
)
from blends.syntax.readers.python import (
    except_clause as python_except_clause,
)
from blends.syntax.readers.python import (
    execution_block as python_execution_block,
)
from blends.syntax.readers.python import (
    expression_statement as python_expression_statement,
)
from blends.syntax.readers.python import (
    finally_clause as python_finally_clause,
)
from blends.syntax.readers.python import (
    for_in_clause as python_for_in_clause,
)
from blends.syntax.readers.python import (
    for_statement as python_for_statement,
)
from blends.syntax.readers.python import (
    function_definition as python_function_definition,
)
from blends.syntax.readers.python import (
    generator_expression as python_generator_expression,
)
from blends.syntax.readers.python import (
    identifier as python_identifier,
)
from blends.syntax.readers.python import (
    if_clause as python_if_clause,
)
from blends.syntax.readers.python import (
    if_statement as python_if_statement,
)
from blends.syntax.readers.python import (
    import_global as python_import_global,
)
from blends.syntax.readers.python import (
    module as python_module,
)
from blends.syntax.readers.python import (
    named_expression as python_named_expression,
)
from blends.syntax.readers.python import (
    not_operator as python_not_operator,
)
from blends.syntax.readers.python import (
    number_literal as python_number_literal,
)
from blends.syntax.readers.python import (
    pair as python_pair,
)
from blends.syntax.readers.python import (
    parameter as python_parameter,
)
from blends.syntax.readers.python import (
    parameters as python_parameters,
)
from blends.syntax.readers.python import (
    parenthesized_expression as python_parenthesized_expression,
)
from blends.syntax.readers.python import (
    raise_statement as python_raise_statement,
)
from blends.syntax.readers.python import (
    reserved_word as python_reserved_word,
)
from blends.syntax.readers.python import (
    return_statement as python_return_statement,
)
from blends.syntax.readers.python import (
    splat_pattern as python_splat_pattern,
)
from blends.syntax.readers.python import (
    string_literal as python_string_literal,
)
from blends.syntax.readers.python import (
    subscript as python_subscript,
)
from blends.syntax.readers.python import (
    try_statement as python_try_statement,
)
from blends.syntax.readers.python import (
    using_statement as python_using_statement,
)
from blends.syntax.readers.python import (
    while_statement as python_while_statement,
)

PYTHON_DISPATCHERS: Dispatcher = {
    "argument_list": python_argument_list.reader,
    "keyword_argument": python_argument.reader,
    "list": python_array.reader,
    "tuple": python_array.reader,
    "tuple_pattern": python_array.reader,
    "pattern_list": python_array.reader,
    "set": python_array.reader,
    "as_pattern": python_as_pattern.reader,
    "assert_statement": python_assert_statement.reader,
    "assignment": python_assignment.reader,
    "attribute": python_attribute.reader,
    "await": python_await_expression.reader,
    "augmented_assignment": python_binary_expression.reader,
    "binary_operator": python_binary_expression.reader,
    "boolean_operator": python_binary_expression.reader,
    "comparison_operator": python_binary_expression.reader,
    "false": python_boolean_literal.reader,
    "true": python_boolean_literal.reader,
    "break_statement": python_break_statement.reader,
    "call": python_call.reader,
    "class_definition": python_class_definition.reader,
    "comment": python_comment.reader,
    "continue_statement": python_continue_statement.reader,
    "pass_statement": python_continue_statement.reader,
    "conditional_expression": python_conditional_expression.reader,
    "decorated_definition": python_decorated_definition.reader,
    "decorator": python_decorator.reader,
    "dictionary": python_dictionary.reader,
    "dictionary_splat": python_dictionary_splat.reader,
    "else_clause": python_else_clause.reader,
    "except_clause": python_except_clause.reader,
    "block": python_execution_block.reader,
    "delete_statement": python_expression_statement.reader,
    "expression_statement": python_expression_statement.reader,
    "global_statement": python_expression_statement.reader,
    "finally_clause": python_finally_clause.reader,
    "for_in_clause": python_for_in_clause.reader,
    "for_statement": python_for_statement.reader,
    "function_definition": python_function_definition.reader,
    "generator_expression": python_generator_expression.reader,
    "identifier": python_identifier.reader,
    "if_clause": python_if_clause.reader,
    "elif_clause": python_if_statement.reader,
    "if_statement": python_if_statement.reader,
    "dotted_name": python_import_global.reader,
    "dictionary_comprehension": python_comprehension.reader,
    "list_comprehension": python_comprehension.reader,
    "set_comprehension": python_comprehension.reader,
    "module": python_module.reader,
    "named_expression": python_named_expression.reader,
    "not_operator": python_not_operator.reader,
    "integer": python_number_literal.reader,
    "float": python_number_literal.reader,
    "pair": python_pair.reader,
    "default_parameter": python_parameter.reader,
    "parameter": python_parameter.reader,
    "typed_parameter": python_parameter.reader,
    "typed_default_parameter": python_parameter.reader,
    "parameters": python_parameters.reader,
    "parenthesized_expression": python_parenthesized_expression.reader,
    "raise_statement": python_raise_statement.reader,
    "in": python_reserved_word.reader,
    "none": python_reserved_word.reader,
    "pass": python_reserved_word.reader,
    "return": python_reserved_word.reader,
    "return_statement": python_return_statement.reader,
    "yield": python_return_statement.reader,
    "list_splat_pattern": python_splat_pattern.reader,
    "dictionary_splat_pattern": python_splat_pattern.reader,
    "expression_list": python_string_literal.reader,
    "list_splat": python_string_literal.reader,
    "slice": python_string_literal.reader,
    "string": python_string_literal.reader,
    "subscript": python_subscript.reader,
    "try_statement": python_try_statement.reader,
    "with_statement": python_using_statement.reader,
    "while_statement": python_while_statement.reader,
}



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.argument_list import (
    build_argument_list_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    childs_id = adj_ast(
        args.ast_graph,
        args.n_id,
    )

    invalid_types = {
        "(",
        ",",
        ")",
    }

    valid_childs = [
        child
        for child in childs_id
        if args.ast_graph.nodes[child]["label_type"] not in invalid_types
    ]

    return build_argument_list_node(args, iter(valid_childs))



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.pair import (
    build_pair_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    c_ids = [_id for _id in adj_ast(graph, args.n_id) if graph.nodes[_id].get("label_type") != "=>"]
    key_id, value_id, *_ = c_ids
    return build_pair_node(args, key_id, value_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
    match_ast_group_d,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    args.syntax_graph.add_node(
        args.n_id,
        label_type="SwitchBody",
    )
    label_ids = match_ast_group_d(graph, args.n_id, "switch_label")
    case_ids = match_ast_group_d(graph, args.n_id, "block")

    for label_id, case_id in zip(label_ids, case_ids, strict=False):
        label_identifier = match_ast(graph, label_id, ":", "case_builtin").get("__0__")
        case_expr = node_to_str(graph, label_identifier) if label_identifier else "Default"

        args.syntax_graph.add_node(
            case_id,
            case_expression=case_expr,
            label_type="SwitchSection",
        )

        args.syntax_graph.add_edge(
            args.n_id,
            case_id,
            label_ast="AST",
        )
        _, *execution_ids, _ = adj_ast(graph, case_id)

        for statement_id in execution_ids:
            args.syntax_graph.add_edge(
                case_id,
                args.generic(args.fork_n_id(statement_id)),
                label_ast="AST",
            )

    return args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.expression_statement import (
    build_expression_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    invalid_types = {
        ";",
    }
    childs_id = adj_ast(graph, args.n_id)
    return build_expression_statement_node(
        args,
        c_ids=(c_id for c_id in childs_id if graph.nodes[c_id]["label_type"] not in invalid_types),
    )



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.for_each_statement import (
    build_for_each_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph

    c_ids = adj_ast(graph, args.n_id)
    min_children_for_foreach_with_iterator = 5
    if len(c_ids) >= min_children_for_foreach_with_iterator:
        iter_item = c_ids[2]
        var_node = c_ids[4]
    else:
        iter_item = c_ids[0]
        var_node = c_ids[-1]

    body_id = graph.nodes[args.n_id].get("label_field_body")

    return build_for_each_statement_node(args, var_node, iter_item, body_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.while_statement import (
    build_while_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    condition_id = n_attrs.get("label_field_condition")
    if not condition_id:
        condition_id = n_attrs.get("label_field_bound_identifier")

    block_id = match_ast_d(graph, args.n_id, "statements")
    if not block_id:
        block_id = adj_ast(graph, args.n_id)[-1]

    return build_while_statement_node(args, block_id, condition_id)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
)
from blends.syntax.builders.file import (
    build_file_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    nodes = graph.nodes

    ignored_types = {"php_tag", "empty_statement"}

    c_ids: set[NId] = set()

    for n_id in adj_ast(graph, args.n_id):
        if nodes[n_id].get("label_type") in ignored_types:
            continue

        if (nodes[n_id].get("label_type") == "expression_statement") and (
            c_id := match_ast(graph, n_id).get("__0__")
        ):
            c_ids.add(c_id)
            continue

        c_ids.add(n_id)

    return build_file_node(args, iter(c_ids))



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.class_decl import (
    build_class_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph

    name = "AnonymousClass"
    name_id = graph.nodes[args.n_id].get("label_field_name") or match_ast_d(
        graph,
        args.n_id,
        "identifier",
    )
    if name_id:
        name = node_to_str(graph, name_id)

    block_id = match_ast_d(graph, args.n_id, "class_body")
    if not block_id:
        block_id = adj_ast(graph, args.n_id)[-1]

    return build_class_node(args, name, block_id, None)



import pytest

from blends.models import SyntaxGraph


@pytest.mark.blends_test_group("syntax_unittesting")
def test_update_node_attribute_valid_key_accepted() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="IfStatement")
    graph.update_node_attribute("1", "condition_id", "2")
    assert graph.nodes["1"]["condition_id"] == "2"


@pytest.mark.blends_test_group("syntax_unittesting")
def test_update_node_attribute_invalid_key_raises() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="IfStatement")
    with pytest.raises(ValueError, match="Invalid node attribute key"):
        graph.update_node_attribute("1", "unknown_key", "bad")


@pytest.mark.blends_test_group("syntax_unittesting")
def test_add_node_invalid_key_raises() -> None:
    graph = SyntaxGraph()
    with pytest.raises(ValueError, match="Invalid node attribute keys"):
        graph.add_node("1", label_type="Return", unknown_key="bad")


@pytest.mark.blends_test_group("syntax_unittesting")
def test_update_node_invalid_key_raises() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="Assignment")
    with pytest.raises(ValueError, match="Invalid node attribute keys"):
        graph.update_node("1", {"unknown_key": "bad"})


@pytest.mark.blends_test_group("syntax_unittesting")
def test_update_node_with_valid_keys_succeeds() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="Return")
    graph.update_node("1", {"value_id": "2", "label_type": "Literal"})
    assert graph.nodes["1"]["value_id"] == "2"
    assert graph.nodes["1"]["label_type"] == "Literal"


@pytest.mark.blends_test_group("syntax_unittesting")
def test_add_node_populates_nodes_by_type_index() -> None:
    graph = SyntaxGraph()
    graph.add_node("1", label_type="IfStatement")
    assert "1" in graph.nodes_by_type["IfStatement"]


@pytest.mark.blends_test_group("syntax_unittesting")
def test_update_node_with_multiple_invalid_keys_raises() -> None:
    graph = SyntaxGraph()
    graph.add_node("1")
    with pytest.raises(ValueError, match="Invalid node attribute keys"):
        graph.update_node("1", {"bad_key_1": "x", "bad_key_2": "y"})



from blends.syntax.models import (
    Dispatcher,
)
from blends.syntax.readers.java import (
    annotation as java_annotation,
)
from blends.syntax.readers.java import (
    annotation_argument_list as java_annotation_argument_list,
)
from blends.syntax.readers.java import (
    argument_list as java_argument_list,
)
from blends.syntax.readers.java import (
    array_access as java_array_access,
)
from blends.syntax.readers.java import (
    array_creation_expression as java_array_creation_expression,
)
from blends.syntax.readers.java import (
    array_node as java_array,
)
from blends.syntax.readers.java import (
    assignment_expression as java_assignment_expression,
)
from blends.syntax.readers.java import (
    binary_expression as java_binary_expression,
)
from blends.syntax.readers.java import (
    boolean_literal as java_boolean_literal,
)
from blends.syntax.readers.java import (
    break_statement as java_break_statement,
)
from blends.syntax.readers.java import (
    cast_expression as java_cast_expression,
)
from blends.syntax.readers.java import (
    catch_clause as java_catch_clause,
)
from blends.syntax.readers.java import (
    catch_declaration as java_catch_declaration,
)
from blends.syntax.readers.java import (
    catch_parameter as java_catch_parameter,
)
from blends.syntax.readers.java import (
    class_body as java_class_body,
)
from blends.syntax.readers.java import (
    class_declaration as java_class_declaration,
)
from blends.syntax.readers.java import (
    comment as java_comment,
)
from blends.syntax.readers.java import (
    continue_statement as java_continue_statement,
)
from blends.syntax.readers.java import (
    declaration_block as java_declaration_block,
)
from blends.syntax.readers.java import (
    do_statement as java_do_statement,
)
from blends.syntax.readers.java import (
    element_value_pair as java_element_value_pair,
)
from blends.syntax.readers.java import (
    enhanced_for_statement as java_enhanced_for_statement,
)
from blends.syntax.readers.java import (
    execution_block as java_execution_block,
)
from blends.syntax.readers.java import (
    expression_statement as java_expression_statement,
)
from blends.syntax.readers.java import (
    field_declaration as java_field_declaration,
)
from blends.syntax.readers.java import (
    finally_clause as java_finally_clause,
)
from blends.syntax.readers.java import (
    for_statement as java_for_statement,
)
from blends.syntax.readers.java import (
    identifier as java_identifier,
)
from blends.syntax.readers.java import (
    if_statement as java_if_statement,
)
from blends.syntax.readers.java import (
    import_declaration as java_import_declaration,
)
from blends.syntax.readers.java import (
    instanceof_expression as java_instanceof_expression,
)
from blends.syntax.readers.java import (
    interface_declaration as java_interface_declaration,
)
from blends.syntax.readers.java import (
    lambda_expression as java_lambda_expression,
)
from blends.syntax.readers.java import (
    method_declaration as java_method_declaration,
)
from blends.syntax.readers.java import (
    method_invocation as java_method_invocation,
)
from blends.syntax.readers.java import (
    modifiers as java_modifiers,
)
from blends.syntax.readers.java import (
    null_literal as java_null_literal,
)
from blends.syntax.readers.java import (
    number_literal as java_number_literal,
)
from blends.syntax.readers.java import (
    object_creation_expression as java_object_creation_expression,
)
from blends.syntax.readers.java import (
    package_declaration as java_package_declaration,
)
from blends.syntax.readers.java import (
    parameter as java_parameter,
)
from blends.syntax.readers.java import (
    parameter_list as java_parameter_list,
)
from blends.syntax.readers.java import (
    parenthesized_expression as java_parenthesized_expression,
)
from blends.syntax.readers.java import (
    program as java_program,
)
from blends.syntax.readers.java import (
    resource_node as java_resource,
)
from blends.syntax.readers.java import (
    return_statement as java_return_statement,
)
from blends.syntax.readers.java import (
    string_literal as java_string_literal,
)
from blends.syntax.readers.java import (
    switch_body as java_switch_body,
)
from blends.syntax.readers.java import (
    switch_section as java_switch_section,
)
from blends.syntax.readers.java import (
    switch_statement as java_switch_statement,
)
from blends.syntax.readers.java import (
    ternary_expression as java_ternary_expression,
)
from blends.syntax.readers.java import (
    this_node as java_this,
)
from blends.syntax.readers.java import (
    throw_statement as java_throw_statement,
)
from blends.syntax.readers.java import (
    try_statement as java_try_statement,
)
from blends.syntax.readers.java import (
    unary_expression as java_unary_expression,
)
from blends.syntax.readers.java import (
    update_expression as java_update_expression,
)
from blends.syntax.readers.java import (
    variable_declaration as java_variable_declaration,
)
from blends.syntax.readers.java import (
    variable_declarator as java_variable_declarator,
)
from blends.syntax.readers.java import (
    while_statement as java_while_statement,
)

JAVA_DISPATCHERS: Dispatcher = {
    "annotation": java_annotation.reader,
    "marker_annotation": java_annotation.reader,
    "annotation_argument_list": java_annotation_argument_list.reader,
    "argument_list": java_argument_list.reader,
    "array_initializer": java_array.reader,
    "element_value_array_initializer": java_array.reader,
    "array_access": java_array_access.reader,
    "array_creation_expression": java_array_creation_expression.reader,
    "assignment_expression": java_assignment_expression.reader,
    "binary_expression": java_binary_expression.reader,
    "true": java_boolean_literal.reader,
    "false": java_boolean_literal.reader,
    "break_statement": java_break_statement.reader,
    "cast_expression": java_cast_expression.reader,
    "catch_clause": java_catch_clause.reader,
    "catch_declaration": java_catch_declaration.reader,
    "catch_formal_parameter": java_catch_parameter.reader,
    "class_body": java_class_body.reader,
    "class_literal": java_identifier.reader,
    "constructor_body": java_class_body.reader,
    "interface_body": java_class_body.reader,
    "class_declaration": java_class_declaration.reader,
    "line_comment": java_comment.reader,
    "block_comment": java_comment.reader,
    "continue_statement": java_continue_statement.reader,
    "declaration_list": java_declaration_block.reader,
    "do_statement": java_do_statement.reader,
    "block": java_execution_block.reader,
    "element_value_pair": java_element_value_pair.reader,
    "enhanced_for_statement": java_enhanced_for_statement.reader,
    "expression_statement": java_expression_statement.reader,
    "field_declaration": java_field_declaration.reader,
    "finally_clause": java_finally_clause.reader,
    "identifier": java_identifier.reader,
    "field_access": java_identifier.reader,
    "scoped_type_identifier": java_identifier.reader,
    "type_identifier": java_identifier.reader,
    "if_statement": java_if_statement.reader,
    "import_declaration": java_import_declaration.reader,
    "instanceof_expression": java_instanceof_expression.reader,
    "interface_declaration": java_interface_declaration.reader,
    "constructor_declaration": java_method_declaration.reader,
    "method_declaration": java_method_declaration.reader,
    "lambda_expression": java_lambda_expression.reader,
    "method_invocation": java_method_invocation.reader,
    "modifiers": java_modifiers.reader,
    "null_literal": java_null_literal.reader,
    "decimal_integer_literal": java_number_literal.reader,
    "integer_literal": java_number_literal.reader,
    "real_literal": java_number_literal.reader,
    "object_creation_expression": java_object_creation_expression.reader,
    "program": java_program.reader,
    "package_declaration": java_package_declaration.reader,
    "for_statement": java_for_statement.reader,
    "formal_parameter": java_parameter.reader,
    "formal_parameters": java_parameter_list.reader,
    "inferred_parameters": java_parameter_list.reader,
    "parenthesized_expression": java_parenthesized_expression.reader,
    "resource": java_resource.reader,
    "resource_specification": java_execution_block.reader,
    "return_statement": java_return_statement.reader,
    "character_literal": java_string_literal.reader,
    "string_literal": java_string_literal.reader,
    "switch_block": java_switch_body.reader,
    "switch_block_statement_group": java_switch_section.reader,
    "switch_expression": java_switch_statement.reader,
    "ternary_expression": java_ternary_expression.reader,
    "this": java_this.reader,
    "throw_statement": java_throw_statement.reader,
    "try_statement": java_try_statement.reader,
    "try_with_resources_statement": java_try_statement.reader,
    "unary_expression": java_unary_expression.reader,
    "update_expression": java_update_expression.reader,
    "constant_declaration": java_variable_declaration.reader,
    "local_variable_declaration": java_variable_declaration.reader,
    "variable_declarator": java_variable_declarator.reader,
    "while_statement": java_while_statement.reader,
}



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.using_statement import (
    build_using_statement_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    block_id = graph.nodes[args.n_id]["label_field_body"]
    if graph.nodes[block_id]["label_type"] == "expression_statement":
        block_id = adj_ast(graph, block_id)[0]

    with_clause = match_ast_d(graph, args.n_id, "with_clause")
    if with_clause:
        with_item = match_ast_d(graph, with_clause, "with_item")
        if with_item:
            with_clause = adj_ast(graph, with_item)[0]

    return build_using_statement_node(args, block_id, with_clause)



from unittest.mock import patch

import pytest

from blends import get_graphs_from_path
from blends.stack.node_kinds import StackGraphNodeKind
from blends.stack.partial_path import (
    PartialPathDatabase,
    PartialPathIndexRequest,
    PartialPathLimits,
    index_partial_paths_for_file,
)
from blends.stack.view import ROOT_INDEX, StackGraphView
from test.stack.test_helpers import STACK_GRAPH_TEST_DATA_DIR

INDEX_MAX_WORK = 50_000


def _load_file(
    test_file: str,
) -> tuple[StackGraphView, PartialPathDatabase]:
    test_file_path = STACK_GRAPH_TEST_DATA_DIR / test_file
    file_path = f"test/data/test_files/stack_graph/{test_file}"

    with patch(
        "blends.ctx.ctx.has_feature_flag",
        side_effect=lambda flag: flag == "StackGraph",
    ):
        graphs = get_graphs_from_path(test_file_path)

    assert graphs.syntax_graph is not None

    view = StackGraphView.from_syntax_graph(graphs.syntax_graph, path=file_path)

    database = PartialPathDatabase()
    index_partial_paths_for_file(
        graphs.syntax_graph,
        request=PartialPathIndexRequest(
            file_path=file_path,
            file_handle=1,
            limits=PartialPathLimits(max_work_per_phase=INDEX_MAX_WORK),
        ),
        database=database,
    )

    return view, database


@pytest.mark.blends_test_group("stack_unittesting")
def test_file_a_has_module_definition_path_from_root() -> None:
    view, _db = _load_file("cross_file_a.py")

    assert view.node_kind[ROOT_INDEX] == StackGraphNodeKind.ROOT.value

    root_successors = [idx for idx, _prec in view.outgoing[ROOT_INDEX]]
    assert len(root_successors) >= 1

    has_pop_for_module = False
    for succ_idx in root_successors:
        kind = view.node_kind[succ_idx]
        if kind == StackGraphNodeKind.POP_SYMBOL.value:
            symbol_id = view.node_symbol_id[succ_idx]
            assert symbol_id is not None
            symbol = view.symbols.id_to_symbol[symbol_id]
            if "cross_file_a" in symbol:
                has_pop_for_module = True

    assert has_pop_for_module, (
        "Expected ROOT to have an outgoing pop_symbol edge containing 'cross_file_a'"
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_file_b_has_import_reference_chain_to_root() -> None:
    view, _db = _load_file("cross_file_b.py")

    assert view.node_kind[ROOT_INDEX] == StackGraphNodeKind.ROOT.value

    push_nodes_to_root = []
    for node_idx in range(len(view.node_kind)):
        kind = view.node_kind[node_idx]
        if kind != StackGraphNodeKind.PUSH_SYMBOL.value:
            continue
        successors = [idx for idx, _prec in view.outgoing[node_idx]]
        if ROOT_INDEX in successors:
            push_nodes_to_root.append(node_idx)

    assert len(push_nodes_to_root) >= 1, "Expected at least one push_symbol node with edge to ROOT"

    pushed_symbols = []
    for node_idx in push_nodes_to_root:
        symbol_id = view.node_symbol_id[node_idx]
        if symbol_id is not None:
            pushed_symbols.append(view.symbols.id_to_symbol[symbol_id])

    assert any("cross_file_a" in sym for sym in pushed_symbols), (
        f"Expected a push toward ROOT with 'cross_file_a', got {pushed_symbols}"
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_partial_paths_are_indexed_for_both_files() -> None:
    _view_a, db_a = _load_file("cross_file_a.py")
    _view_b, db_b = _load_file("cross_file_b.py")

    record_a = db_a.get_file(1)
    record_b = db_b.get_file(1)

    assert record_a is not None
    assert record_b is not None
    assert len(record_a.paths) > 0, "File A should have partial paths"
    assert len(record_b.paths) > 0, "File B should have partial paths"


@pytest.mark.blends_test_group("stack_unittesting")
def test_definition_and_reference_paths_have_compatible_symbols() -> None:
    view_a, db_a = _load_file("cross_file_a.py")
    view_b, db_b = _load_file("cross_file_b.py")

    record_a = db_a.get_file(1)
    record_b = db_b.get_file(1)
    assert record_a is not None
    assert record_b is not None

    def_precondition_symbols: set[str] = set()
    for path in record_a.paths:
        if path.start_node_index == ROOT_INDEX:
            for scoped_sym in path.symbol_stack_precondition.symbols:
                symbol = view_a.symbols.id_to_symbol[scoped_sym.symbol_id]
                def_precondition_symbols.add(symbol)

    ref_postcondition_symbols: set[str] = set()
    for path in record_b.paths:
        if path.end_node_index == ROOT_INDEX:
            for scoped_sym in path.symbol_stack_postcondition.symbols:
                symbol = view_b.symbols.id_to_symbol[scoped_sym.symbol_id]
                ref_postcondition_symbols.add(symbol)

    shared_symbols = def_precondition_symbols & ref_postcondition_symbols
    assert len(shared_symbols) > 0, (
        f"Expected shared symbols between definition preconditions {def_precondition_symbols} "
        f"and reference postconditions {ref_postcondition_symbols}"
    )



from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_while_statement_node(args: SyntaxGraphArgs, block: NId, conditional: NId | None) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        block_id=block,
        label_type="WhileStatement",
    )

    args.syntax_graph.add_edge(
        args.n_id,
        args.generic(args.fork_n_id(block)),
        label_ast="AST",
    )

    if conditional:
        args.syntax_graph.update_node_attribute(args.n_id, "condition_id", conditional)
        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(conditional)),
            label_ast="AST",
        )

    return args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast_d,
)
from blends.syntax.builders.annotation import (
    build_annotation_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    args_id = None
    annotation_name = "@"

    constructor_id = match_ast_d(args.ast_graph, args.n_id, "constructor_invocation")
    if (
        constructor_id
        and (children := adj_ast(args.ast_graph, constructor_id))
        and len(children) > 0
    ):
        annotation_name += node_to_str(args.ast_graph, children[0])
        args_id = children[1]

    return build_annotation_node(args, annotation_name, args_id)



from blends.models import (
    NId,
)
from blends.syntax.builders.if_statement import (
    build_if_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    condition_id = n_attrs["label_field_condition"]
    true_id = n_attrs["label_field_body"]

    return build_if_node(args, condition_id, true_id)



from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_return_node(args: SyntaxGraphArgs, value_id: NId | None) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        label_type="Return",
    )
    if value_id:
        args.syntax_graph.update_node_attribute(args.n_id, "value_id", value_id)
        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(value_id)),
            label_ast="AST",
        )

    return args.n_id



from blends.ctx import ctx
from blends.models import (
    NId,
)
from blends.stack.edges import (
    Edge,
    add_edge,
)
from blends.stack.node_helpers import (
    import_binding_node_attributes,
    pop_scoped_symbol_node_attributes,
    pop_symbol_node_attributes,
)
from blends.syntax.builders.import_reference_chain import (
    build_import_reference_chain_to_root,
)
from blends.syntax.builders.utils import (
    get_next_binding_precedence,
    get_next_synthetic_node_id,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def _add_class_attr_unscoped_definition(
    args: SyntaxGraphArgs,
    variable_name: str,
    class_scope_nid: NId,
    precedence: int,
) -> None:
    unscoped_def_id = get_next_synthetic_node_id(args)
    args.syntax_graph.add_node(
        unscoped_def_id,
        label_type="SyntheticClassAttrDef",
        **pop_symbol_node_attributes(
            symbol=variable_name, precedence=precedence, is_definition=True
        ),
    )
    ast_node = args.ast_graph.nodes.get(args.n_id, {})
    if (label_l := ast_node.get("label_l")) is not None:
        args.syntax_graph.update_node_attribute(unscoped_def_id, "label_l", label_l)
        args.syntax_graph.update_node_attribute(
            unscoped_def_id, "label_c", ast_node.get("label_c", "0")
        )
    add_edge(
        args.syntax_graph,
        Edge(source=class_scope_nid, sink=unscoped_def_id, precedence=precedence),
    )


def _wire_variable_stack_graph_node(
    args: SyntaxGraphArgs,
    variable_name: str,
    import_module_stem: str | None,
    import_source_symbol: str | None,
) -> None:
    precedence = get_next_binding_precedence(args)
    if import_module_stem is not None and import_source_symbol is not None:
        args.syntax_graph.update_node(
            args.n_id,
            import_binding_node_attributes(
                symbol=variable_name,
                precedence=precedence,
                import_module_stem=import_module_stem,
                import_source_symbol=import_source_symbol,
            ),
        )
    else:
        args.syntax_graph.update_node(
            args.n_id,
            pop_scoped_symbol_node_attributes(
                symbol=variable_name,
                precedence=precedence,
            ),
        )
    scope_stack = args.metadata.setdefault("scope_stack", [])
    if scope_stack and (parent_scope := scope_stack[-1]):
        add_edge(
            args.syntax_graph,
            Edge(source=parent_scope, sink=args.n_id, precedence=precedence),
        )
    if import_module_stem is not None and import_source_symbol is not None:
        build_import_reference_chain_to_root(args, import_module_stem, args.n_id)
    class_member_scope_stack = args.metadata.get("class_member_scope_stack", [])
    if (
        class_member_scope_stack
        and scope_stack
        and class_member_scope_stack[-1][0] == scope_stack[-1]
    ):
        _add_class_attr_unscoped_definition(
            args, variable_name, class_member_scope_stack[-1][0], precedence
        )


def build_variable_declaration_node(  # noqa: PLR0913
    args: SyntaxGraphArgs,
    variable_name: str,
    variable_type: str | None,
    value_id: NId | None,
    var_id: NId | None = None,
    access_modifier: str | None = None,
    import_module_stem: str | None = None,
    import_source_symbol: str | None = None,
) -> NId:
    args.syntax_graph.add_node(
        args.n_id,
        label_type="VariableDeclaration",
        variable=variable_name,
    )

    if variable_type:
        args.syntax_graph.update_node_attribute(args.n_id, "variable_type", variable_type)

    if value_id:
        args.syntax_graph.update_node_attribute(args.n_id, "value_id", value_id)

        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(value_id)),
            label_ast="AST",
        )
    if var_id:
        args.syntax_graph.update_node_attribute(args.n_id, "variable_id", var_id)

        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(var_id)),
            label_ast="AST",
        )

    if access_modifier:
        args.syntax_graph.update_node_attribute(args.n_id, "access_modifier", access_modifier)

    if ctx.has_feature_flag("StackGraph"):
        _wire_variable_stack_graph_node(
            args, variable_name, import_module_stem, import_source_symbol
        )

    return args.n_id



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.unary_expression import (
    build_unary_expression_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    childs = adj_ast(graph, args.n_id)
    operator = graph.nodes[childs[1]]["label_text"]
    return build_unary_expression_node(args, operator, childs[0])



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_group_d,
)
from blends.syntax.builders.switch_body import (
    build_switch_body_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    case_ids = match_ast_group_d(graph, args.n_id, "switch_block_statement_group")
    return build_switch_body_node(args, case_ids, None)



from blends.models import NId
from blends.query import adj_ast, match_ast_group_d
from blends.syntax.builders.method_declaration import (
    DirectChildren,
    build_method_declaration_node,
)
from blends.syntax.models import SyntaxGraphArgs


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph

    parameters_list = match_ast_group_d(
        graph,
        args.n_id,
        "lambda_function_type_parameters",
    )

    body_id = None
    for child_id in adj_ast(graph, args.n_id):
        child_attrs = graph.nodes[child_id]
        if child_attrs.get("label_type") in {"statements", "function_body"}:
            body_id = child_id
            break

    direct: DirectChildren = {}
    if body_id:
        direct["block_id"] = body_id
    if parameters_list:
        direct["parameters_id"] = parameters_list[0]

    return build_method_declaration_node(args, None, direct, {})



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_group_d,
)
from blends.syntax.builders.jsx_element import (
    build_jsx_element_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    childs_with_attrs = match_ast_group_d(graph, args.n_id, "jsx_attribute")
    childs_with_attrs.append(graph.nodes[args.n_id]["label_field_name"])

    return build_jsx_element_node(args, childs_with_attrs)



from blends.syntax.models import (
    Dispatcher,
)
from blends.syntax.readers.swift import (
    array_node as swift_array_node,
)
from blends.syntax.readers.swift import (
    as_expression as swift_as_expression,
)
from blends.syntax.readers.swift import (
    assignment as swift_assignment,
)
from blends.syntax.readers.swift import (
    availability_condition as swift_availability_condition,
)
from blends.syntax.readers.swift import (
    await_expression as swift_await_expression,
)
from blends.syntax.readers.swift import (
    binary_expression as swift_binary_expression,
)
from blends.syntax.readers.swift import (
    boolean_literal as swift_boolean_literal,
)
from blends.syntax.readers.swift import (
    call_expression as swift_call_expresion,
)
from blends.syntax.readers.swift import (
    call_suffix as swift_call_suffix,
)
from blends.syntax.readers.swift import (
    class_body as swift_class_body,
)
from blends.syntax.readers.swift import (
    class_declaration as swift_class_declaration,
)
from blends.syntax.readers.swift import (
    comment as swift_comment,
)
from blends.syntax.readers.swift import (
    control_transfer as swift_control_transfer,
)
from blends.syntax.readers.swift import (
    dictionary_literal as swift_dictionary_literal,
)
from blends.syntax.readers.swift import (
    do_statement as swift_do_statement,
)
from blends.syntax.readers.swift import (
    enum_entry as swift_enum_entry,
)
from blends.syntax.readers.swift import (
    expression_statement as swift_expression_statement,
)
from blends.syntax.readers.swift import (
    for_statement as swift_for_statement,
)
from blends.syntax.readers.swift import (
    function_body as swift_function_body,
)
from blends.syntax.readers.swift import (
    function_declaration as swift_function_declaration,
)
from blends.syntax.readers.swift import (
    identifier as swift_identifier,
)
from blends.syntax.readers.swift import (
    if_nil_expression as swift_if_nil_expression,
)
from blends.syntax.readers.swift import (
    if_statement as swift_if_statement,
)
from blends.syntax.readers.swift import (
    import_statement as swift_import_statement,
)
from blends.syntax.readers.swift import (
    integer_literal as swift_integer_literal,
)
from blends.syntax.readers.swift import (
    lambda_function as swift_lambda_function,
)
from blends.syntax.readers.swift import (
    navigation_expression as swift_navigation_expression,
)
from blends.syntax.readers.swift import (
    navigation_suffix as swift_navigation_suffix,
)
from blends.syntax.readers.swift import (
    nil as swift_nil,
)
from blends.syntax.readers.swift import (
    parameter as swift_parameter,
)
from blends.syntax.readers.swift import (
    prefix_expression as swift_prefix_expression,
)
from blends.syntax.readers.swift import (
    property_declaration as swift_property_declaration,
)
from blends.syntax.readers.swift import (
    reserved_word as swift_reserved_word,
)
from blends.syntax.readers.swift import (
    source_file as swift_source_file,
)
from blends.syntax.readers.swift import (
    statements as swift_statements,
)
from blends.syntax.readers.swift import (
    string_literal as swift_string_literal,
)
from blends.syntax.readers.swift import (
    switch_entry as swift_switch_entry,
)
from blends.syntax.readers.swift import (
    switch_statement as swift_switch_statement,
)
from blends.syntax.readers.swift import (
    ternary_expression as swift_ternary_expression,
)
from blends.syntax.readers.swift import (
    try_expression as swift_try_expression,
)
from blends.syntax.readers.swift import (
    update_expression as swift_update_expression,
)
from blends.syntax.readers.swift import (
    value_argument as swift_value_argument,
)
from blends.syntax.readers.swift import (
    value_arguments as swift_argument_list,
)
from blends.syntax.readers.swift import (
    while_statement as swift_while_statement,
)

SWIFT_DISPATCHERS: Dispatcher = {
    "lambda_literal": swift_lambda_function.reader,
    "call_suffix": swift_call_suffix.reader,
    "assignment": swift_assignment.reader,
    "availability_condition": swift_availability_condition.reader,
    "await_expression": swift_await_expression.reader,
    "value_arguments": swift_argument_list.reader,
    "additive_expression": swift_binary_expression.reader,
    "bitwise_operation": swift_binary_expression.reader,
    "comparison_expression": swift_binary_expression.reader,
    "conjunction_expression": swift_binary_expression.reader,
    "disjunction_expression": swift_binary_expression.reader,
    "equality_expression": swift_binary_expression.reader,
    "infix_expression": swift_binary_expression.reader,
    "multiplicative_expression": swift_binary_expression.reader,
    "boolean_literal": swift_boolean_literal.reader,
    "call_expression": swift_call_expresion.reader,
    "class_body": swift_class_body.reader,
    "enum_class_body": swift_class_body.reader,
    "protocol_body": swift_class_body.reader,
    "class_declaration": swift_class_declaration.reader,
    "protocol_declaration": swift_class_declaration.reader,
    "comment": swift_comment.reader,
    "directive": swift_comment.reader,
    "multiline_comment": swift_comment.reader,
    "control_transfer_statement": swift_control_transfer.reader,
    "enum_entry": swift_enum_entry.reader,
    "do_statement": swift_do_statement.reader,
    "directly_assignable_expression": swift_expression_statement.reader,
    "pattern": swift_expression_statement.reader,
    "for_statement": swift_for_statement.reader,
    "computed_property": swift_function_body.reader,
    "function_body": swift_function_body.reader,
    "function_declaration": swift_function_declaration.reader,
    "init_declaration": swift_function_declaration.reader,
    "protocol_function_declaration": swift_function_declaration.reader,
    "array_literal": swift_array_node.reader,
    "as_expression": swift_as_expression.reader,
    "constructor_expression": swift_identifier.reader,
    "identifier": swift_identifier.reader,
    "open_start_range_expression": swift_identifier.reader,
    "postfix_expression": swift_update_expression.reader,
    "self_expression": swift_identifier.reader,
    "simple_identifier": swift_identifier.reader,
    "switch_pattern": swift_identifier.reader,
    "nil_coalescing_expression": swift_if_nil_expression.reader,
    "guard_statement": swift_if_statement.reader,
    "if_statement": swift_if_statement.reader,
    "import_declaration": swift_import_statement.reader,
    "integer_literal": swift_integer_literal.reader,
    "navigation_expression": swift_navigation_expression.reader,
    "navigation_suffix": swift_navigation_suffix.reader,
    "nil": swift_nil.reader,
    "parameter": swift_parameter.reader,
    "prefix_expression": swift_prefix_expression.reader,
    "property_declaration": swift_property_declaration.reader,
    "protocol_property_declaration": swift_property_declaration.reader,
    "typealias_declaration": swift_property_declaration.reader,
    "source_file": swift_source_file.reader,
    "super_expression": swift_reserved_word.reader,
    "statements": swift_statements.reader,
    "switch_entry": swift_switch_entry.reader,
    "switch_statement": swift_switch_statement.reader,
    "dictionary_literal": swift_dictionary_literal.reader,
    "enum_type_parameters": swift_string_literal.reader,
    "line_string_literal": swift_string_literal.reader,
    "multi_line_string_literal": swift_string_literal.reader,
    "tuple_expression": swift_string_literal.reader,
    "user_type": swift_string_literal.reader,
    "ternary_expression": swift_ternary_expression.reader,
    "try_expression": swift_try_expression.reader,
    "value_argument": swift_value_argument.reader,
    "while_statement": swift_while_statement.reader,
    "repeat_while_statement": swift_while_statement.reader,
}



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.pair import (
    build_pair_node,
)
from blends.syntax.builders.symbol_lookup import (
    build_symbol_lookup_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    key_id = n_attrs.get("label_field_key")
    value_id = n_attrs.get("label_field_value")
    if value_id and key_id:
        if graph.nodes[value_id]["label_type"] == "block_node":
            value_id = adj_ast(graph, value_id)[-1]
        return build_pair_node(args, key_id, value_id)

    symbol = node_to_str(graph, args.n_id)
    return build_symbol_lookup_node(args, symbol)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
    match_ast,
    match_ast_d,
    match_ast_group_d,
)
from blends.syntax.builders.method_declaration import (
    DirectChildren,
    ListChildren,
    build_method_declaration_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    if declaration_n_id := match_ast_d(graph, args.n_id, "identifier"):
        graph.nodes[args.n_id]["label_l"] = graph.nodes[declaration_n_id]["label_l"]
    name = node_to_str(graph, n_attrs["label_field_name"])
    block_id = n_attrs.get("label_field_body")
    n_access_mod = [
        adj[0]
        for node in adj_ast(graph, args.n_id, label_type="modifier")
        if (adj := adj_ast(graph, node))
    ]

    parameters_id = n_attrs["label_field_parameters"]
    if "__0__" not in match_ast(graph, parameters_id, "(", ")"):
        parameters_id = None

    attributes_ids = match_ast_group_d(graph, args.n_id, "attribute_list")

    direct: DirectChildren = {}
    if block_id:
        direct["block_id"] = block_id
    if parameters_id:
        direct["parameters_id"] = parameters_id

    list_children: ListChildren = {}
    if attributes_ids:
        list_children["modifiers_ids"] = attributes_ids

    access_modifiers = (
        " ".join(graph.nodes[i].get("label_text", "") for i in n_access_mod)
        if n_access_mod
        else None
    )

    return build_method_declaration_node(args, name, direct, list_children, access_modifiers)



from blends.models import (
    NId,
)
from blends.query import (
    adj_ast,
)
from blends.syntax.builders.object import (
    build_object_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    valid_parameters = {"class_body"}
    graph = args.ast_graph
    c_ids = adj_ast(graph, args.n_id)
    name = ""
    for child in c_ids:
        if graph.nodes[child]["label_type"] in {
            "delegation_specifiers",
            "type_identifier",
        }:
            name = node_to_str(graph, child)
    return build_object_node(
        args,
        c_ids=(_id for _id in c_ids if graph.nodes[_id]["label_type"] in valid_parameters),
        name=name,
    )



from blends.models import (
    NId,
)
from blends.syntax.builders.class_decl import (
    build_class_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import (
    node_to_str,
)


def reader(args: SyntaxGraphArgs) -> NId:
    class_node = args.ast_graph.nodes[args.n_id]
    name_id = class_node.get("label_field_name")
    name = "DefaultClass"
    if name_id:
        name = node_to_str(args.ast_graph, name_id)
    block_id = class_node.get("label_field_body")

    attributes_id = class_node.get("label_field_type_parameters")
    attributes_list = [attributes_id] if attributes_id else []

    return build_class_node(args, name, block_id, attributes_list)



import pytest

from blends.models import SyntaxGraph
from blends.stack.attributes import STACK_GRAPH_KIND
from blends.stack.node_kinds import StackGraphNodeKind
from blends.stack.partial_path import (
    PartialPath,
    PartialPathDb,
    PartialPathResolutionError,
    PartialScopedSymbol,
    PartialScopeStack,
    PartialSymbolStack,
    PerFilePartialPathStats,
    ScopeStackVariable,
    StackGraphViewMetadata,
    SymbolStackVariable,
)
from blends.stack.partial_path_database import PartialPathFileRecord
from blends.stack.view import StackGraphView


def _path(
    *,
    start: int,
    end: int,
    symbol_ids: tuple[int, ...],
    has_variable: bool = False,
) -> PartialPath:
    symbols = tuple(
        PartialScopedSymbol(symbol_id=symbol_id, scopes=None) for symbol_id in symbol_ids
    )
    variable = SymbolStackVariable.initial() if has_variable else None
    symbol_stack = PartialSymbolStack(symbols=symbols, variable=variable)
    return PartialPath(
        start_node_index=start,
        end_node_index=end,
        symbol_stack_precondition=symbol_stack,
        symbol_stack_postcondition=symbol_stack,
        scope_stack_precondition=PartialScopeStack.empty(),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )


def _record(file_handle: int, paths: tuple[PartialPath, ...]) -> PartialPathFileRecord:
    metadata = StackGraphViewMetadata(
        file_handles=("test.py",),
        symbol_id_to_string=("a", "b", "c", "d"),
    )
    return PartialPathFileRecord(
        file_handle=file_handle,
        file_path="test.py",
        view_snapshot=metadata,
        paths=paths,
        stats=PerFilePartialPathStats(),
    )


def _view_with_node(node_id: str, kind: str) -> StackGraphView:
    graph = SyntaxGraph()
    graph.add_node(node_id)
    graph.update_node_attribute(node_id, STACK_GRAPH_KIND, kind)
    return StackGraphView.from_syntax_graph(graph, path="test.py")


@pytest.mark.blends_test_group("stack_unittesting")
def test_start_node_index_returns_paths_for_node() -> None:
    db = PartialPathDb()
    paths = (
        _path(start=2, end=3, symbol_ids=(0,)),
        _path(start=2, end=4, symbol_ids=(1,)),
        _path(start=5, end=6, symbol_ids=(2,)),
    )
    db.add_file(_record(1, paths))

    candidates = db.get_candidates(
        file_handle=1,
        end_node_index=2,
        symbol_stack=PartialSymbolStack.empty(),
    )

    assert candidates == ((1, 0), (1, 1))


@pytest.mark.blends_test_group("stack_unittesting")
def test_root_exact_match_returns_expected_paths() -> None:
    db = PartialPathDb()
    paths = (
        _path(start=0, end=7, symbol_ids=(0,)),
        _path(start=0, end=8, symbol_ids=(1,)),
    )
    db.add_file(_record(1, paths))

    candidates = db.get_candidates(
        file_handle=1,
        end_node_index=0,
        symbol_stack=PartialSymbolStack(
            symbols=(PartialScopedSymbol(symbol_id=0, scopes=None),),
            variable=None,
        ),
    )

    assert candidates == ((1, 0),)


@pytest.mark.blends_test_group("stack_unittesting")
def test_root_variable_query_includes_longer_and_shorter_variable_prefixes() -> None:
    db = PartialPathDb()
    paths = (
        _path(start=0, end=10, symbol_ids=(0,), has_variable=False),
        _path(start=0, end=11, symbol_ids=(0, 1), has_variable=False),
        _path(start=0, end=12, symbol_ids=(0, 1), has_variable=True),
        _path(start=0, end=13, symbol_ids=(), has_variable=True),
    )
    db.add_file(_record(1, paths))

    query_stack = PartialSymbolStack(
        symbols=(PartialScopedSymbol(symbol_id=0, scopes=None),),
        variable=SymbolStackVariable.initial(),
    )
    candidates = db.get_candidates(
        file_handle=1,
        end_node_index=0,
        symbol_stack=query_stack,
    )

    assert (1, 0) in candidates
    assert (1, 1) in candidates
    assert (1, 2) in candidates
    assert (1, 3) in candidates


@pytest.mark.blends_test_group("stack_unittesting")
def test_root_incompatible_stack_returns_empty() -> None:
    db = PartialPathDb()
    paths = (_path(start=0, end=20, symbol_ids=(0,)),)
    db.add_file(_record(1, paths))

    candidates = db.get_candidates(
        file_handle=1,
        end_node_index=0,
        symbol_stack=PartialSymbolStack(
            symbols=(PartialScopedSymbol(symbol_id=1, scopes=None),),
            variable=None,
        ),
    )

    assert candidates == ()


@pytest.mark.blends_test_group("stack_unittesting")
def test_root_candidates_are_deterministic() -> None:
    paths = (
        _path(start=0, end=30, symbol_ids=(0,), has_variable=False),
        _path(start=0, end=31, symbol_ids=(0,), has_variable=True),
    )
    record = _record(1, paths)

    db_one = PartialPathDb()
    db_one.add_file(record)
    db_two = PartialPathDb()
    db_two.add_file(record)

    query_stack = PartialSymbolStack(
        symbols=(PartialScopedSymbol(symbol_id=0, scopes=None),),
        variable=None,
    )
    candidates_one = db_one.get_candidates(
        file_handle=1,
        end_node_index=0,
        symbol_stack=query_stack,
    )
    candidates_two = db_two.get_candidates(
        file_handle=1,
        end_node_index=0,
        symbol_stack=query_stack,
    )

    assert candidates_one == candidates_two


@pytest.mark.blends_test_group("stack_unittesting")
def test_integration_non_root_candidates_concatenate() -> None:
    db = PartialPathDb()
    expected_end_index = 3
    left = _path(start=1, end=2, symbol_ids=())
    right = _path(start=2, end=expected_end_index, symbol_ids=())
    db.add_file(_record(1, (left, right)))

    candidates = db.get_candidates(
        file_handle=1,
        end_node_index=2,
        symbol_stack=PartialSymbolStack.empty(),
    )

    assert candidates == ((1, 1),)

    view = _view_with_node("2", StackGraphNodeKind.SCOPE.value)
    candidate = db.get_path(candidates[0])
    assert candidate is not None
    concatenated = left.concatenate(view, candidate)

    assert concatenated.start_node_index == 1
    assert concatenated.end_node_index == expected_end_index


@pytest.mark.blends_test_group("stack_unittesting")
def test_integration_root_candidates_concatenate_success_and_failure() -> None:
    db = PartialPathDb()
    expected_end_index = 4
    expected_bad_end_index = 5
    symbol = PartialScopedSymbol(symbol_id=0, scopes=None)
    left = PartialPath(
        start_node_index=0,
        end_node_index=0,
        symbol_stack_precondition=PartialSymbolStack.empty(),
        symbol_stack_postcondition=PartialSymbolStack(
            symbols=(symbol,),
            variable=SymbolStackVariable.initial(),
        ),
        scope_stack_precondition=PartialScopeStack.empty(),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )
    right_ok = PartialPath(
        start_node_index=0,
        end_node_index=expected_end_index,
        symbol_stack_precondition=PartialSymbolStack(symbols=(symbol,), variable=None),
        symbol_stack_postcondition=PartialSymbolStack(symbols=(), variable=None),
        scope_stack_precondition=PartialScopeStack.empty(),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )
    right_bad = PartialPath(
        start_node_index=0,
        end_node_index=expected_bad_end_index,
        symbol_stack_precondition=PartialSymbolStack(symbols=(symbol,), variable=None),
        symbol_stack_postcondition=PartialSymbolStack(symbols=(), variable=None),
        scope_stack_precondition=PartialScopeStack(scopes=(1,), variable=None),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )
    db.add_file(_record(1, (left, right_ok, right_bad)))

    candidates = db.get_candidates(
        file_handle=1,
        end_node_index=0,
        symbol_stack=left.symbol_stack_postcondition,
    )

    assert (1, 1) in candidates
    assert (1, 2) in candidates

    view = StackGraphView.from_syntax_graph(SyntaxGraph(), path="test.py")
    candidate_paths = [db.get_path(candidate) for candidate in candidates]
    assert all(path is not None for path in candidate_paths)
    resolved_candidates = [path for path in candidate_paths if path is not None]
    ok_candidate = next(
        path for path in resolved_candidates if path.end_node_index == expected_end_index
    )
    assert left.concatenate(view, ok_candidate).end_node_index == expected_end_index
    bad_candidate = next(
        path for path in resolved_candidates if path.end_node_index == expected_bad_end_index
    )
    with pytest.raises(PartialPathResolutionError):
        left.concatenate(view, bad_candidate)

    filtered_candidates = db.get_candidates(
        file_handle=1,
        end_node_index=0,
        symbol_stack=left.symbol_stack_postcondition,
        scope_stack=left.scope_stack_postcondition,
    )
    assert (1, 1) in filtered_candidates
    assert (1, 2) not in filtered_candidates


@pytest.mark.blends_test_group("stack_unittesting")
def test_scope_filtering_rejects_incompatible_candidates() -> None:
    db = PartialPathDb()
    symbol = PartialScopedSymbol(symbol_id=0, scopes=None)
    path_scopes = (
        PartialPath(
            start_node_index=0,
            end_node_index=7,
            symbol_stack_precondition=PartialSymbolStack(symbols=(symbol,), variable=None),
            symbol_stack_postcondition=PartialSymbolStack(symbols=(), variable=None),
            scope_stack_precondition=PartialScopeStack(scopes=(99,), variable=None),
            scope_stack_postcondition=PartialScopeStack.empty(),
        ),
        PartialPath(
            start_node_index=0,
            end_node_index=8,
            symbol_stack_precondition=PartialSymbolStack(symbols=(symbol,), variable=None),
            symbol_stack_postcondition=PartialSymbolStack(symbols=(), variable=None),
            scope_stack_precondition=PartialScopeStack(scopes=(6,), variable=None),
            scope_stack_postcondition=PartialScopeStack.empty(),
        ),
        PartialPath(
            start_node_index=0,
            end_node_index=9,
            symbol_stack_precondition=PartialSymbolStack(symbols=(symbol,), variable=None),
            symbol_stack_postcondition=PartialSymbolStack(symbols=(), variable=None),
            scope_stack_precondition=PartialScopeStack(scopes=(5, 6), variable=None),
            scope_stack_postcondition=PartialScopeStack.empty(),
        ),
    )
    db.add_file(_record(1, path_scopes))

    query_symbol = PartialSymbolStack(symbols=(symbol,), variable=SymbolStackVariable.initial())
    assert (
        db.get_candidates(
            file_handle=1,
            end_node_index=0,
            symbol_stack=query_symbol,
            scope_stack=PartialScopeStack.empty(),
        )
        == ()
    )
    assert (
        db.get_candidates(
            file_handle=1,
            end_node_index=0,
            symbol_stack=query_symbol,
            scope_stack=PartialScopeStack(scopes=(5,), variable=None),
        )
        == ()
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_scope_filtering_passes_compatible_candidates() -> None:
    """Passes empty vs empty, (5,)+var vs (5,)+var, and variable absorption in both directions."""
    db = PartialPathDb()
    symbol = PartialScopedSymbol(symbol_id=0, scopes=None)
    path_empty_a = PartialPath(
        start_node_index=0,
        end_node_index=7,
        symbol_stack_precondition=PartialSymbolStack(symbols=(symbol,), variable=None),
        symbol_stack_postcondition=PartialSymbolStack(symbols=(), variable=None),
        scope_stack_precondition=PartialScopeStack.empty(),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )
    path_empty_b = PartialPath(
        start_node_index=0,
        end_node_index=8,
        symbol_stack_precondition=PartialSymbolStack(symbols=(symbol,), variable=None),
        symbol_stack_postcondition=PartialSymbolStack(symbols=(), variable=None),
        scope_stack_precondition=PartialScopeStack.empty(),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )
    path_var = PartialPath(
        start_node_index=0,
        end_node_index=9,
        symbol_stack_precondition=PartialSymbolStack(symbols=(symbol,), variable=None),
        symbol_stack_postcondition=PartialSymbolStack(symbols=(), variable=None),
        scope_stack_precondition=PartialScopeStack(
            scopes=(5,), variable=ScopeStackVariable.initial()
        ),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )
    path_longer = PartialPath(
        start_node_index=0,
        end_node_index=10,
        symbol_stack_precondition=PartialSymbolStack(symbols=(symbol,), variable=None),
        symbol_stack_postcondition=PartialSymbolStack(symbols=(), variable=None),
        scope_stack_precondition=PartialScopeStack(scopes=(5, 6, 7), variable=None),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )
    db.add_file(_record(1, (path_empty_a, path_empty_b, path_var, path_longer)))

    query_symbol = PartialSymbolStack(symbols=(symbol,), variable=SymbolStackVariable.initial())
    empty = db.get_candidates(
        file_handle=1,
        end_node_index=0,
        symbol_stack=query_symbol,
        scope_stack=PartialScopeStack.empty(),
    )
    expected_empty_count = 2
    assert len(empty) == expected_empty_count
    assert (1, 0) in empty
    assert (1, 1) in empty

    var_query = db.get_candidates(
        file_handle=1,
        end_node_index=0,
        symbol_stack=query_symbol,
        scope_stack=PartialScopeStack(scopes=(5,), variable=ScopeStackVariable.initial()),
    )
    expected_var_match_count = 2
    assert len(var_query) == expected_var_match_count
    assert (1, 2) in var_query
    assert (1, 3) in var_query

    longer_query = db.get_candidates(
        file_handle=1,
        end_node_index=0,
        symbol_stack=query_symbol,
        scope_stack=PartialScopeStack(scopes=(5, 6), variable=None),
    )
    assert len(longer_query) == 1
    assert (1, 2) in longer_query


@pytest.mark.blends_test_group("stack_unittesting")
def test_scope_filtering_none_returns_all() -> None:
    db = PartialPathDb()
    symbol = PartialScopedSymbol(symbol_id=0, scopes=None)
    path_a = PartialPath(
        start_node_index=0,
        end_node_index=7,
        symbol_stack_precondition=PartialSymbolStack(symbols=(symbol,), variable=None),
        symbol_stack_postcondition=PartialSymbolStack(symbols=(), variable=None),
        scope_stack_precondition=PartialScopeStack(scopes=(99,), variable=None),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )
    path_b = PartialPath(
        start_node_index=0,
        end_node_index=8,
        symbol_stack_precondition=PartialSymbolStack(symbols=(symbol,), variable=None),
        symbol_stack_postcondition=PartialSymbolStack(symbols=(), variable=None),
        scope_stack_precondition=PartialScopeStack.empty(),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )
    db.add_file(_record(1, (path_a, path_b)))

    candidates = db.get_candidates(
        file_handle=1,
        end_node_index=0,
        symbol_stack=PartialSymbolStack(symbols=(symbol,), variable=SymbolStackVariable.initial()),
        scope_stack=None,
    )

    assert (1, 0) in candidates
    assert (1, 1) in candidates


@pytest.mark.blends_test_group("stack_unittesting")
def test_two_files_overlapping_root_symbols_both_returned() -> None:
    fh1, fh2 = 1, 2
    db = PartialPathDb()
    db.add_file(_record(fh1, (_path(start=0, end=10, symbol_ids=(0,)),)))
    db.add_file(_record(fh2, (_path(start=0, end=20, symbol_ids=(0,)),)))

    query_stack = PartialSymbolStack(
        symbols=(PartialScopedSymbol(symbol_id=0, scopes=None),),
        variable=None,
    )
    candidates = db.get_candidates(
        file_handle=fh1,
        end_node_index=0,
        symbol_stack=query_stack,
    )

    assert (fh1, 0) in candidates
    assert (fh2, 0) in candidates
    assert fh1 in db.root_exact_index
    assert fh2 in db.root_exact_index


@pytest.mark.blends_test_group("stack_unittesting")
def test_drop_file_removes_only_target_file_entries() -> None:
    db = PartialPathDb()
    db.add_file(
        _record(
            1,
            (
                _path(start=0, end=10, symbol_ids=(0,)),
                _path(start=2, end=3, symbol_ids=(1,)),
            ),
        )
    )
    db.add_file(
        _record(
            2,
            (
                _path(start=0, end=20, symbol_ids=(0,)),
                _path(start=2, end=30, symbol_ids=(1,)),
            ),
        )
    )

    db.drop_file(1)

    query_stack = PartialSymbolStack(
        symbols=(PartialScopedSymbol(symbol_id=0, scopes=None),),
        variable=None,
    )
    root_candidates = db.get_candidates(
        file_handle=2,
        end_node_index=0,
        symbol_stack=query_stack,
    )
    assert (1, 0) not in root_candidates
    assert (2, 0) in root_candidates
    assert 1 not in db.root_exact_index

    assert (
        db.get_candidates(file_handle=1, end_node_index=2, symbol_stack=PartialSymbolStack.empty())
        == ()
    )
    assert (2, 1) in db.get_candidates(
        file_handle=2, end_node_index=2, symbol_stack=PartialSymbolStack.empty()
    )


@pytest.mark.blends_test_group("stack_unittesting")
def test_scope_filtering_reduces_candidates_considered() -> None:
    db = PartialPathDb()
    symbol = PartialScopedSymbol(symbol_id=0, scopes=None)
    query_scope = PartialScopeStack.empty()
    query_symbol = PartialSymbolStack(symbols=(symbol,), variable=SymbolStackVariable.initial())

    compatible = PartialPath(
        start_node_index=0,
        end_node_index=10,
        symbol_stack_precondition=PartialSymbolStack(symbols=(symbol,), variable=None),
        symbol_stack_postcondition=PartialSymbolStack(symbols=(), variable=None),
        scope_stack_precondition=PartialScopeStack.empty(),
        scope_stack_postcondition=PartialScopeStack.empty(),
    )
    incompatible_paths = tuple(
        PartialPath(
            start_node_index=0,
            end_node_index=20 + i,
            symbol_stack_precondition=PartialSymbolStack(symbols=(symbol,), variable=None),
            symbol_stack_postcondition=PartialSymbolStack(symbols=(), variable=None),
            scope_stack_precondition=PartialScopeStack(scopes=(100 + i,), variable=None),
            scope_stack_postcondition=PartialScopeStack.empty(),
        )
        for i in range(4)
    )
    db.add_file(_record(1, (compatible, *incompatible_paths)))

    unfiltered = db.get_candidates(
        file_handle=1,
        end_node_index=0,
        symbol_stack=query_symbol,
        scope_stack=None,
    )
    filtered = db.get_candidates(
        file_handle=1,
        end_node_index=0,
        symbol_stack=query_symbol,
        scope_stack=query_scope,
    )

    num_incompatible = 4
    expected_total = 1 + num_incompatible
    assert len(unfiltered) == expected_total
    assert len(filtered) == 1
    avoided_concatenations = len(unfiltered) - len(filtered)
    assert avoided_concatenations == num_incompatible



from blends.models import (
    NId,
)
from blends.syntax.builders.element_access import (
    build_element_access_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    n_attrs = args.ast_graph.nodes[args.n_id]
    arguments_id = n_attrs["label_field_subscript"]
    expr_id = n_attrs["label_field_value"]
    return build_element_access_node(args, expr_id, arguments_id)



from blends.syntax.node_attributes import (
    STACK_GRAPH_IMPORT_MODULE_STEM,
    STACK_GRAPH_IMPORT_SOURCE_SYMBOL,
    STACK_GRAPH_IS_DEFINITION,
    STACK_GRAPH_IS_EXPORTED,
    STACK_GRAPH_IS_IMPORT_BINDING,
    STACK_GRAPH_IS_REFERENCE,
    STACK_GRAPH_KIND,
    STACK_GRAPH_PRECEDENCE,
    STACK_GRAPH_SCOPE,
    STACK_GRAPH_SCOPE_TYPE,
    STACK_GRAPH_SYMBOL,
)

__all__ = [
    "STACK_GRAPH_IMPORT_MODULE_STEM",
    "STACK_GRAPH_IMPORT_SOURCE_SYMBOL",
    "STACK_GRAPH_IS_DEFINITION",
    "STACK_GRAPH_IS_EXPORTED",
    "STACK_GRAPH_IS_IMPORT_BINDING",
    "STACK_GRAPH_IS_REFERENCE",
    "STACK_GRAPH_KIND",
    "STACK_GRAPH_PRECEDENCE",
    "STACK_GRAPH_SCOPE",
    "STACK_GRAPH_SCOPE_TYPE",
    "STACK_GRAPH_SYMBOL",
]



import pytest

from blends.models import SyntaxGraph
from blends.stack.attributes import (
    STACK_GRAPH_IS_DEFINITION,
    STACK_GRAPH_IS_REFERENCE,
    STACK_GRAPH_KIND,
    STACK_GRAPH_SYMBOL,
)
from blends.stack.criteria import is_definition_endpoint
from blends.stack.forward_stitcher import (
    ForwardStitcher,
    ForwardStitcherConfig,
)
from blends.stack.node_kinds import StackGraphNodeKind
from blends.stack.partial_path import PartialPathDb
from blends.stack.view import StackGraphView
from test.stack.test_helpers import make_path, view_for_graph
from test.stack.test_helpers import record as make_record


def _view_with_reference_join_definition() -> StackGraphView:
    graph = SyntaxGraph()
    graph.add_node("ref")
    graph.update_node_attribute("ref", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("ref", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("ref", STACK_GRAPH_IS_REFERENCE, value=True)

    graph.add_node("join")
    graph.update_node_attribute("join", STACK_GRAPH_KIND, StackGraphNodeKind.SCOPE.value)

    graph.add_node("def")
    graph.update_node_attribute("def", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("def", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("def", STACK_GRAPH_IS_DEFINITION, value=True)

    return view_for_graph(graph)


@pytest.mark.blends_test_group("stack_unittesting")
def test_forward_stitcher_respects_phase_work_limit() -> None:
    view = _view_with_reference_join_definition()
    ref_index = view.nid_to_index["ref"]
    def_index = view.nid_to_index["def"]

    db = PartialPathDb()
    db.add_file(make_record((make_path(ref_index, def_index),)))

    stitcher = ForwardStitcher.from_references(
        view=view,
        file_handle=1,
        db=db,
        reference_nodes=(ref_index,),
        config=ForwardStitcherConfig(max_work_per_phase=0),
    )

    result = stitcher.process_next_phase()

    assert result.complete_paths == ()
    assert result.stats.candidates_considered == 0
    assert stitcher.previous_phase_paths() != ()
    assert not stitcher.is_complete()


@pytest.mark.blends_test_group("stack_unittesting")
def test_forward_stitcher_finds_complete_path_single_phase() -> None:
    view = _view_with_reference_join_definition()
    ref_index = view.nid_to_index["ref"]
    def_index = view.nid_to_index["def"]

    db = PartialPathDb()
    db.add_file(make_record((make_path(ref_index, def_index),)))

    stitcher = ForwardStitcher.from_references(
        view=view,
        file_handle=1,
        db=db,
        reference_nodes=(ref_index,),
        config=ForwardStitcherConfig(),
    )

    result = stitcher.process_next_phase()

    assert len(result.complete_paths) == 1
    assert is_definition_endpoint(view, result.complete_paths[0].end_node_index)


@pytest.mark.blends_test_group("stack_unittesting")
def test_forward_stitcher_stitches_across_two_paths() -> None:
    view = _view_with_reference_join_definition()
    ref_index = view.nid_to_index["ref"]
    join_index = view.nid_to_index["join"]
    def_index = view.nid_to_index["def"]

    db = PartialPathDb()
    db.add_file(make_record((make_path(ref_index, join_index), make_path(join_index, def_index))))

    stitcher = ForwardStitcher.from_references(
        view=view,
        file_handle=1,
        db=db,
        reference_nodes=(ref_index,),
        config=ForwardStitcherConfig(),
    )

    first = stitcher.process_next_phase()
    assert first.complete_paths == ()

    second = stitcher.process_next_phase()
    assert len(second.complete_paths) == 1
    assert is_definition_endpoint(view, second.complete_paths[0].end_node_index)


@pytest.mark.blends_test_group("stack_unittesting")
def test_forward_stitcher_stitches_via_root() -> None:
    view = _view_with_reference_join_definition()
    ref_index = view.nid_to_index["ref"]
    def_index = view.nid_to_index["def"]

    db = PartialPathDb()
    db.add_file(make_record((make_path(ref_index, 0), make_path(0, def_index))))

    stitcher = ForwardStitcher.from_references(
        view=view,
        file_handle=1,
        db=db,
        reference_nodes=(ref_index,),
        config=ForwardStitcherConfig(),
    )

    stitcher.process_next_phase()
    result = stitcher.process_next_phase()

    assert len(result.complete_paths) == 1
    assert is_definition_endpoint(view, result.complete_paths[0].end_node_index)


def _view_with_two_definitions() -> StackGraphView:
    graph = SyntaxGraph()

    graph.add_node("ref")
    graph.update_node_attribute("ref", STACK_GRAPH_KIND, StackGraphNodeKind.PUSH_SYMBOL.value)
    graph.update_node_attribute("ref", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("ref", STACK_GRAPH_IS_REFERENCE, value=True)

    graph.add_node("join")
    graph.update_node_attribute("join", STACK_GRAPH_KIND, StackGraphNodeKind.SCOPE.value)

    graph.add_node("def_a")
    graph.update_node_attribute("def_a", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("def_a", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("def_a", STACK_GRAPH_IS_DEFINITION, value=True)

    graph.add_node("def_b")
    graph.update_node_attribute("def_b", STACK_GRAPH_KIND, StackGraphNodeKind.POP_SYMBOL.value)
    graph.update_node_attribute("def_b", STACK_GRAPH_SYMBOL, "x")
    graph.update_node_attribute("def_b", STACK_GRAPH_IS_DEFINITION, value=True)

    return view_for_graph(graph)


@pytest.mark.blends_test_group("stack_unittesting")
def test_complete_paths_are_not_re_queued_for_further_extension() -> None:
    view = _view_with_reference_join_definition()
    ref_index = view.nid_to_index["ref"]
    join_index = view.nid_to_index["join"]
    def_index = view.nid_to_index["def"]

    db = PartialPathDb()
    db.add_file(
        make_record(
            (
                make_path(ref_index, join_index),
                make_path(join_index, def_index),
                make_path(def_index, join_index),
            )
        )
    )

    stitcher = ForwardStitcher.from_references(
        view=view,
        file_handle=1,
        db=db,
        reference_nodes=(ref_index,),
        config=ForwardStitcherConfig(),
    )

    phase_one = stitcher.process_next_phase()
    assert phase_one.complete_paths == ()

    phase_two = stitcher.process_next_phase()
    assert len(phase_two.complete_paths) == 1
    assert is_definition_endpoint(view, phase_two.complete_paths[0].end_node_index)
    assert stitcher.is_complete()
    assert phase_two.stats.candidates_considered == 2  # noqa: PLR2004
    assert phase_two.stats.complete_paths_found == 1


@pytest.mark.blends_test_group("stack_unittesting")
def test_multi_definition_correctness_is_preserved() -> None:
    view = _view_with_two_definitions()
    ref_index = view.nid_to_index["ref"]
    join_index = view.nid_to_index["join"]
    def_a_index = view.nid_to_index["def_a"]
    def_b_index = view.nid_to_index["def_b"]

    db = PartialPathDb()
    db.add_file(
        make_record(
            (
                make_path(ref_index, join_index),
                make_path(join_index, def_a_index),
                make_path(join_index, def_b_index),
            ),
            symbol_id_to_string=("x",),
        )
    )

    stitcher = ForwardStitcher.from_references(
        view=view,
        file_handle=1,
        db=db,
        reference_nodes=(ref_index,),
        config=ForwardStitcherConfig(),
    )

    phase_one = stitcher.process_next_phase()
    assert phase_one.complete_paths == ()

    phase_two = stitcher.process_next_phase()
    assert len(phase_two.complete_paths) == 2  # noqa: PLR2004

    found_endpoints = {path.end_node_index for path in phase_two.complete_paths}
    assert def_a_index in found_endpoints
    assert def_b_index in found_endpoints

    assert phase_two.stats.complete_paths_found == 2  # noqa: PLR2004



from blends.models import (
    NId,
)
from blends.query import (
    match_ast_d,
)
from blends.syntax.builders.parameter import (
    build_parameter_node,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)
from blends.utilities.text_nodes import node_to_str


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    name_n_id = n_attrs.get("label_field_name")
    var_name = None
    if (name_n_id is not None) and (name_n_id := match_ast_d(graph, name_n_id, "name")):
        var_name = graph.nodes[name_n_id].get("label_text", "")

    type_n_id = n_attrs.get("label_field_type")
    var_type = None
    if type_n_id is not None:
        var_type = node_to_str(graph, type_n_id)

    return build_parameter_node(args=args, variable=var_name, variable_type=var_type, value_id=None)



from blends.models import NId
from blends.syntax.builders.unary_expression import build_unary_expression_node
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def reader(args: SyntaxGraphArgs) -> NId:
    graph = args.ast_graph
    n_attrs = graph.nodes[args.n_id]
    operator = (
        graph.nodes[operator_id]["label_text"]
        if (operator_id := n_attrs.get("label_field_operator"))
        else ""
    )
    return build_unary_expression_node(args, operator, NId(n_attrs["label_field_operand"]))



from blends.syntax.models import (
    Dispatcher,
)
from blends.syntax.readers.php import (
    anon_method_declaration as php_anon_method_decl,
)
from blends.syntax.readers.php import (
    argument_list as php_argument_list,
)
from blends.syntax.readers.php import (
    array_creation as php_array_creation,
)
from blends.syntax.readers.php import (
    arrow as php_arrow,
)
from blends.syntax.readers.php import (
    assignment as php_assignment,
)
from blends.syntax.readers.php import (
    binary_expression as php_binary_expression,
)
from blends.syntax.readers.php import (
    boolean_literal as php_boolean_literal,
)
from blends.syntax.readers.php import (
    break_statement as php_break_statement,
)
from blends.syntax.readers.php import (
    case_statement as php_case_statement,
)
from blends.syntax.readers.php import (
    catch_clause as php_catch_clause,
)
from blends.syntax.readers.php import (
    class_declaration as php_class_declaration,
)
from blends.syntax.readers.php import (
    comment as php_comment,
)
from blends.syntax.readers.php import (
    const_declaration as php_const_declaration,
)
from blends.syntax.readers.php import (
    constant_access as php_constant_access,
)
from blends.syntax.readers.php import (
    continue_statement as php_continue_statement,
)
from blends.syntax.readers.php import (
    default_statement as php_default_statement,
)
from blends.syntax.readers.php import (
    echo_arguments as php_echo_arguments,
)
from blends.syntax.readers.php import (
    execution_block as php_execution_block,
)
from blends.syntax.readers.php import (
    expression_statement as php_expression_statement,
)
from blends.syntax.readers.php import (
    for_each_statement as php_for_each_statement,
)
from blends.syntax.readers.php import (
    for_statement as php_for_statement,
)
from blends.syntax.readers.php import (
    identifier as php_identifier,
)
from blends.syntax.readers.php import (
    if_statement as php_if_statement,
)
from blends.syntax.readers.php import (
    import_statement as php_import_statement,
)
from blends.syntax.readers.php import (
    inner_use_import as php_inner_use_import,
)
from blends.syntax.readers.php import (
    member_access as php_member_access,
)
from blends.syntax.readers.php import (
    method_declaration as php_method_declaration,
)
from blends.syntax.readers.php import (
    method_invocation as php_method_invocation,
)
from blends.syntax.readers.php import (
    namespace_definition as php_namespace_definition,
)
from blends.syntax.readers.php import (
    number_literal as php_number_literal,
)
from blends.syntax.readers.php import (
    object_creation as php_object_creation,
)
from blends.syntax.readers.php import (
    pair as php_pair,
)
from blends.syntax.readers.php import (
    parameter as php_parameter,
)
from blends.syntax.readers.php import (
    parameter_list as php_parameter_list,
)
from blends.syntax.readers.php import (
    print_statement as php_print_statement,
)
from blends.syntax.readers.php import (
    program as php_program,
)
from blends.syntax.readers.php import (
    property_access as php_property_access,
)
from blends.syntax.readers.php import (
    reserved_word as php_reserved_word,
)
from blends.syntax.readers.php import (
    return_statement as php_return_statement,
)
from blends.syntax.readers.php import (
    scoped_invocation as php_scoped_invocation,
)
from blends.syntax.readers.php import (
    string_literal as php_string_literal,
)
from blends.syntax.readers.php import (
    subscript as php_subscript,
)
from blends.syntax.readers.php import (
    switch_block as php_switch_block,
)
from blends.syntax.readers.php import (
    switch_statement as php_switch_statement,
)
from blends.syntax.readers.php import (
    text_interpolation as php_text_interpolation,
)
from blends.syntax.readers.php import (
    throw_statement as php_throw_statement,
)
from blends.syntax.readers.php import (
    try_statement as php_try_statement,
)
from blends.syntax.readers.php import (
    unary_op_expression as php_unary_operation,
)
from blends.syntax.readers.php import (
    update_expression as php_update_expression,
)
from blends.syntax.readers.php import (
    use_import as php_use_import,
)
from blends.syntax.readers.php import (
    variable_declaration as php_variable_declaration,
)
from blends.syntax.readers.php import (
    variable_name as php_variable_name,
)
from blends.syntax.readers.php import (
    while_statement as php_while_statement,
)

PHP_DISPATCHERS: Dispatcher = {
    "=>": php_arrow.reader,
    "anonymous_function_creation_expression": php_anon_method_decl.reader,
    "arguments": php_argument_list.reader,
    "array_creation_expression": php_array_creation.reader,
    "assignment_expression": php_assignment.reader,
    "augmented_assignment_expression": php_binary_expression.reader,
    "binary_expression": php_binary_expression.reader,
    "boolean": php_boolean_literal.reader,
    "break_statement": php_break_statement.reader,
    "case_statement": php_case_statement.reader,
    "catch_clause": php_catch_clause.reader,
    "class_constant_access_expression": php_constant_access.reader,
    "class_declaration": php_class_declaration.reader,
    "comment": php_comment.reader,
    "compound_statement": php_execution_block.reader,
    "const_declaration": php_const_declaration.reader,
    "continue_statement": php_continue_statement.reader,
    "declaration_list": php_execution_block.reader,
    "default_statement": php_default_statement.reader,
    "dynamic_variable_name": php_expression_statement.reader,
    "echo": php_echo_arguments.reader,
    "echo_statement": php_print_statement.reader,
    "encapsed_string": php_string_literal.reader,
    "exit_statement": php_print_statement.reader,
    "float": php_number_literal.reader,
    "foreach_statement": php_for_each_statement.reader,
    "for_statement": php_for_statement.reader,
    "formal_parameters": php_parameter_list.reader,
    "function_definition": php_method_declaration.reader,
    "function_call_expression": php_method_invocation.reader,
    "if_statement": php_if_statement.reader,
    "else_if_clause": php_if_statement.reader,
    "include_expression": php_import_statement.reader,
    "include_once_expression": php_import_statement.reader,
    "integer": php_number_literal.reader,
    "member_access_expression": php_member_access.reader,
    "member_call_expression": php_method_invocation.reader,
    "method_declaration": php_method_declaration.reader,
    "name": php_identifier.reader,
    "namespace_definition": php_namespace_definition.reader,
    "namespace_use_declaration": php_use_import.reader,
    "null": php_reserved_word.reader,
    "object_creation_expression": php_object_creation.reader,
    "pair": php_pair.reader,
    "parent": php_identifier.reader,
    "print_intrinsic": php_print_statement.reader,
    "program": php_program.reader,
    "property_declaration": php_variable_declaration.reader,
    "require_expression": php_import_statement.reader,
    "require_once_expression": php_import_statement.reader,
    "return_statement": php_return_statement.reader,
    "scoped_call_expression": php_scoped_invocation.reader,
    "scoped_property_access_expression": php_property_access.reader,
    "self": php_identifier.reader,
    "simple_parameter": php_parameter.reader,
    "string": php_string_literal.reader,
    "subscript_expression": php_subscript.reader,
    "switch_block": php_switch_block.reader,
    "switch_statement": php_switch_statement.reader,
    "text_interpolation": php_text_interpolation.reader,
    "throw_expression": php_throw_statement.reader,
    "try_statement": php_try_statement.reader,
    "unary_op_expression": php_unary_operation.reader,
    "update_expression": php_update_expression.reader,
    "use_declaration": php_inner_use_import.reader,
    "variable_name": php_variable_name.reader,
    "while_statement": php_while_statement.reader,
}



from blends.models import (
    NId,
)
from blends.syntax.models import (
    SyntaxGraphArgs,
)


def build_annotation_node(args: SyntaxGraphArgs, attr_name: str, al_id: NId | None) -> NId:
    n_args = {
        "name": attr_name,
        "label_type": "Annotation",
    }

    if al_id:
        n_args.update(
            {
                "arguments_id": al_id,
            },
        )
    args.syntax_graph.add_node(args.n_id, **n_args)

    if al_id:
        args.syntax_graph.add_edge(
            args.n_id,
            args.generic(args.fork_n_id(al_id)),
            label_ast="AST",
        )

    return args.n_id
